#ifndef DIJKSTRA_HPP_
#define DIJKSTRA_HPP_

#include "PriorityQueue.hpp"
#include <vector>


template< typename graph_slot, typename PriorityQueueType  >
class Dijkstra{
  typedef graph_slot Graph;
  typedef typename Graph::Node Node;
  typedef typename Graph::Edge Edge;
  typedef typename Edge::weight_type Weight;
  typedef typename Edge::meta_weight MetaWeight;  //Meta-Information for the edge-weight
public:
  /* Construction */
  Dijkstra( const Graph & graph_ )
    : graph( graph_ ),
      priority_queue( graph_ )
  {}

  /* Destruction */
  ~Dijkstra(){}

  bool search( unsigned int source, unsigned int target, Weight & result ) {
    priority_queue.clear();
    priority_queue.push( source, Weight( 0 ) );   //initial insert of source node
    while( !priority_queue.empty() ){
      unsigned int current = priority_queue.getMin();
      Weight current_weight = priority_queue.getMinKey();
      priority_queue.deleteMin();

      if( current == target ){
        result = current_weight;
        return true;
      }

      for( unsigned int eid = graph.edgeBegin( current ); eid != graph.edgeEnd( current ); ++eid ){
        const Edge & edge = graph.getEdge( eid );
        if( edge.forward ){
          Weight new_weight = edge.weight + current_weight;
          if( !priority_queue.isReached( edge.target ) ){
            priority_queue.push( edge.target, new_weight );
          } else {
            if( priority_queue.getCurrentKey( edge.target ) > new_weight ){
              priority_queue.decreaseKey( edge.target, new_weight );
            }
          }
        }
      }
    }
    return false; //Node not found
  }

  Weight distance(unsigned int nid ) const{
    return priority_queue.getCurrentKey( nid );
  }

private:
  const Graph & graph;
  PriorityQueueType priority_queue;
};


#endif /* DIJKSTRA_HPP_ */
