/*
 * queryStaticDijkstra.cpp
 *
 *  Created on: May 17, 2011
 *      Author: kobitzsch
 */

#include <unistd.h>

//Graph, Node and Edges
#include "KStaticGraph.hpp"
#include "BasicNode.h"
#include "Edge.hpp"

//Query
#include "Dijkstra.hpp"
#include "PriorityQueue.hpp"

//Timing
#include "timer.h"

#include <iostream>
#include <fstream>
#include <string>

#include <cstdlib>

int main( int argc, char ** args ){
	if( argc == 1 ){
		std::cout << "Usage: " << args[0] << std::endl;
		std::cout << "\t-g contracted graph" << std::endl;
		return 0;
	}

	typedef BasicNode Node;
	typedef DirectedIntegerWeightedEdge Edge;
	typedef Edge::weight_type Weight;
	typedef KStaticGraph<Node,Edge> Graph;

	Graph graph;

	int c;
	while( (c = getopt( argc, args, "g:o:q:n:" ) ) != -1 ){
		switch( c ){
		case 'g':{
				std::ifstream tmp( optarg, std::ios::in | std::ios::binary );
				graph.deserialize( tmp );
				tmp.close();
			}break;
		}
	}

	typedef PriorityQueue<Graph, unsigned int, unsigned int, std::numeric_limits<unsigned int> > PriorityQueueT;
	Dijkstra<Graph,PriorityQueueT> query( graph );

	ClockCounterTimer timer( 2.0 * 1000000 );	//ersetz das durch die Maschine, auf der gerechnet wird
	//ClockCounterTimer timer( 1.8660 * 1000000 );	//ersetz das durch die Maschine, auf der gerechnet wird
	//unsigned int numQueries = 1;
	//unsigned int numQueries = 15;
	unsigned int numQueries = 100; //wahrscheinlich mehr fuer die challange bewertung
	double sum = 0;
  unsigned int sources[] = {3558682,4017064,1594836,3488048,3725078,2337475,474409,2667690,1071821,2789390,2375739,4081846,4355882,3314046,3166827,1969138,4196395};
  unsigned int targets[] = {2802450,1221656,2967857,1551428,3723874,2102442,3372837,2020947,302074,3846488,3014559,319047,4130347,38098,4209046,3829085,3295599};
	for( unsigned int i = 0; i < numQueries; ++i ){
		//unsigned int source = 3725078, target = 3725081, weight;
		//unsigned int source = 3725078, target = 3725090, weight;
		//unsigned int source = 3725078, target = 3723874, weight;
		//unsigned int source = 3314046, target = 38098, weight;
		//unsigned int source = 3166827, target = 4209046, weight;
    //unsigned int source = 1969138, target = 3829085, weight;
    //unsigned int source = 4196395, target = 3295599, weight;
    //unsigned int source = 3287467, target = 2724823, weight;
    //unsigned int source = 1594836, target = 2967857, weight;
		//unsigned int source = sources[i], target = targets[i], weight;
		unsigned int source = rand() % graph.numberOfNodes(), target = rand() % graph.numberOfNodes(), weight;
		timer.start();
    //cout << source << "," << target << endl;
		bool connected = query.search( ( source ), ( target ), weight );
	  double t = timer.stop();
    sum += t;
    std::cout << "Weight: " << weight << " (" << source << "->" << target << "), time: " << t << std::endl;
	}
	std::cout << "Avg time: " << sum / numQueries / 1000 << " seconds." << std::endl;

	return 0;
}
