/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  class Element {
    public:
      id_slot id;
      key_slot key;
      Element* next;
      Element* prev;

      Element(id_slot i = UINT_MAX, key_slot k = UINT_MAX)
      : id(i)
      , key(k)
      , next(NULL)
      , prev(NULL)
      {}
  };

  unsigned int _size;
  Element* b;
  Element** p;
  Element* n;
  unsigned int pSize;
  unsigned int nCounter;
  unsigned int maxWeight;
  unsigned int mask;
  unsigned int curPos;

  inline void insertBefore(register Element* e, register Element* n) {
    register Element* nn = e->prev;
#ifdef DEBUG
    cout << "nn: " << nn << endl;
#endif
    n->next = e;
    n->prev = nn;
    nn->next = n;
    e->prev = n;
  }

  inline void insertAfter(register Element* e, register Element* n) {
    register Element* nn = e->next;
    n->prev = e;
    n->next = nn;
    nn->prev = n;
    e->next = n;
  }

  inline void cutOut(Element* e) {
    register Element* p = e->prev;
    register Element* n = e->next;
    n->prev = p;
    p->next = n;
  }

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(262144)
  , mask(maxWeight - 1)
  {
    cout << "go" << endl;
    p = new Element *[graph.numberOfNodes()];
    n = new Element[graph.numberOfNodes()];

    //for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
    //  unsigned int newWeight = graph.getEdge(eid).getWeight();
    //  //unsigned int newWeight = 2000;

    //  if (newWeight > maxWeight) {
    //    maxWeight = newWeight;
    //  }
    //}

    //maxWeight += 2;
    //cout << "maxWeight: " << maxWeight << endl;

    b = new Element[maxWeight];

    for (int i = 0; i <= maxWeight; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      b[i].id = UINT_MAX;
      b[i].key = UINT_MAX;
#ifdef DEBUG
      cout << "b[" << i << "] = " << &b[i] << endl;
#endif
    }

    //for (int i = 0; i < pSize; i++) {
    //  p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);
    //}
  }

  inline size_t size() const{
    return _size;
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  void printList(Element* e) {
    Element *x = e->next;
    cout << "----------" << endl;
    cout << e << "(id:" << e->id << ",key:" << e->key << "," << e->prev << "," << e->next << endl;
    while (x != e) {
      cout << x << "(id:" << x->id << ",key:" << x->key << "," << x->prev << "," << x->next << endl;
      x = x->next;
    }
    cout << "----------" << endl;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    unsigned int pos = key & mask;
    register Element* e = &n[nCounter++];
    p[id] = e;
    e->id = id;
    e->key = key;
    ++_size;

    insertAfter(&b[pos], e);
  }

  void deleteMin(){
#ifdef DEBUG/*{{{*/
      cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    --_size;

    cutOut(b[curPos].next);
  }

  inline const id_slot & getMin() {
    while (b[curPos].next == &b[curPos]) {
      curPos = (curPos + 1) & mask;
    }
    return b[curPos].next->id;
  }

  inline const key_slot & getMinKey() const{
    return b[curPos].next->key;
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id]->key << endl;
#endif/*}}}*/
    return p[id]->key;
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != NULL) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != NULL;
    //return p[id].key != UINT_MAX;
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    unsigned int pos = new_key & mask;
    Element* e = p[id];
    e->key = new_key;

    cutOut(e);
    insertAfter(&b[pos], e);
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    _size = 0;
    nCounter = 0;
    curPos = 0;

    for (register unsigned int i = 0; i <= maxWeight; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      b[i].next = &b[i];
      b[i].prev = &b[i];
    }

    memset(p, 0, pSize * sizeof(Element*));
    //for (int i = 0; i < pSize; i++) {
    //  //n[i].key = UINT_MAX;
    //  p[i] = NULL;

    //  //if (p[i] != NULL) {
    //  //  delete p[i];
    //  //}
    //  //p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);

    //  //p[i]->id = i;
    //  //p[i]->next = NULL;
    //  //p[i]->prev = NULL;
    //}
  }
};


#endif /* BINARY_HEAP_HPP_ */
