/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>
#include <limits>

using namespace std;

#define DBS 512

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    Element* next;
  };

  struct IKElement {
    id_slot id;
    key_slot key;
  };

  const unsigned int pSize;
  unsigned int maxWeight;
  unsigned int mask;
  unsigned int nCounter;
  unsigned int curPos;
  unsigned int curKey;
  unsigned int _size;
  unsigned int dbPos;

  Element* b;
  Element* n;
  key_slot* p;
  id_slot* delBuffer;

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(0)
  {
    cout << "go" << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    p = new key_slot[pSize];
    n = new Element[pSize];
    b = new Element[maxWeight];
    delBuffer = new id_slot[DBS];
    //p = (key_slot*) (((unsigned long) (new char[(pSize)*sizeof(key_slot) + 256]) + 256) & ~0xFF);
    //n = (Element*) (((unsigned long) (new char[(pSize)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //b = (Element*) (((unsigned long) (new char[(maxWeight)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //aligned_storage<262144*sizeof(Element), alignment_of<Element>::value > buf;
    //b = reinterpret_cast<Element&>(buf);
  }

  inline size_t size() const{
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/
    //_size++;

    p[id] = key;
    unsigned int pos = key & mask;

    if (b[pos].id == numeric_limits<id_slot>::max()) {
      b[pos].id = id;
    } else {
      Element* e = &n[nCounter++];
      e->id = id;
      e->next = b[pos].next;
      b[pos].next = e;
    }
  }

  void deleteMin(){
#ifdef DEBUG/*{{{*/
      cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
  }

  inline const id_slot getMin() {
    if (dbPos != 0) {
      return delBuffer[--dbPos];
    }

    while (b[curPos].id == numeric_limits<id_slot>::max()) {
      curPos = (curPos + 1) & mask;
      curKey++;
    }

    id_slot curId = b[curPos].id;

    if (b[curPos].next != (Element*) -1) {
      Element* e = b[curPos].next;
      while (e != (Element*) -1) {
        delBuffer[dbPos++] = e->id;
        e = e->next;
      }
      b[curPos].next = (Element*) -1;
    }

    b[curPos].id = numeric_limits<id_slot>::max();

    return curId;
  }

  inline const key_slot & getMinKey() const{
    return curKey;
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != numeric_limits<key_slot>::max()) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != numeric_limits<key_slot>::max();
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    register unsigned int pos = new_key & mask;
    register unsigned int oldPos = p[id] & mask;

    p[id] = new_key;
    register Element* e = NULL;

    if (b[oldPos].id == id) {
      if (b[oldPos].next == (Element*) -1) {
        b[oldPos].id = numeric_limits<id_slot>::max();
      } else {
        e = b[oldPos].next;
        b[oldPos].id = e->id;
        b[oldPos].next = e->next;
      }
    } else {
      Element* f = &b[oldPos];

      while (f->next->id != id) {
        f = f->next;
      }

      e = f->next;
      f->next = e->next;
    }

    if (b[pos].id != numeric_limits<id_slot>::max()) {
      if (e == NULL) {
        e = &n[nCounter++];
      }
      e->id = id;
      e->next = b[pos].next;
      b[pos].next = e;
    } else {
      b[pos].id = id;
    }
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    nCounter = 0;
    curPos = 0;
    curKey = 0;
    //_size = 0;
    dbPos = 0;

    memset(b, 0xFF, maxWeight * sizeof(Element)); // bad style...
    //for (register unsigned int i = 0; i < maxWeight; i++) {
    //  b[i].next = NULL;
    //  b[i].key = UINT_MAX;
    //  //b[i].next = &b[i];
    //}

    memset(p, 0xFF, pSize * sizeof(key_slot));
    //fill_n(p, pSize, UINT_MAX);

    memset(delBuffer, 0xFF, DBS * sizeof(id_slot));
  }
};


#endif /* BINARY_HEAP_HPP_ */
