/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

#define DEBUG

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>
#include <limits>

using namespace std;

#define MAX 0xFFFFFF

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    //id_slot id;
    //unsigned int next;
    //Element* next;
    unsigned char id[3];
    unsigned char next[3];
  };

  static void copy3(unsigned char to[3], unsigned char from[3]) {
    memcpy(to, from, 3);
  }

  static void to3(unsigned char to[3], unsigned int from) {
    memcpy(to, &from, 3); // TODO: make not endian-dependant
    //to[2] = (from << 8) >> 16;
    //to[1] = (from << 16) >> 16;
    //to[0] = (from << 24) >> 16;
  }

  static unsigned int from3(unsigned char from[3]) {
    return from[2] << 16 | from[1] << 8 | from[0];
  }

  const unsigned int pSize;
  unsigned int maxWeight;
  unsigned int mask;
  unsigned int nCounter;
  unsigned int curPos;
  unsigned int curKey;

  Element* b;
  Element* n;
  key_slot* p;

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(0)
  {
    cout << "go" << endl;
    cout << sizeof(id_slot) << endl;
    cout << sizeof(Element*) << endl;
    cout << pSize << endl;

    unsigned char to[3];
    to3(to, 0);
    cout << from3(to) << endl;
    unsigned char copy[3];
    copy3(copy, to);
    cout << from3(copy) << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    p = new key_slot[pSize];
    n = new Element[pSize];
    b = new Element[maxWeight];
    //p = (key_slot*) (((unsigned long) (new char[(pSize)*sizeof(key_slot) + 256]) + 256) & ~0xFF);
    //n = (Element*) (((unsigned long) (new char[(pSize)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //b = (Element*) (((unsigned long) (new char[(maxWeight)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //aligned_storage<262144*sizeof(Element), alignment_of<Element>::value > buf;
    //b = reinterpret_cast<Element&>(buf);
  }

  size_t size() const{
  }

  bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    p[id] = key;
    unsigned int pos = key & mask;

    if (from3(b[pos].id) == MAX) {
      cout << "a" << endl;
      to3(b[pos].id, id);
      cout << (unsigned int) b[pos].id[0] << endl;
      cout << (unsigned int) b[pos].id[1] << endl;
      cout << (unsigned int) b[pos].id[2] << endl;
      cout << from3(b[pos].id) << endl;
    } else {
      cout << "b" << endl;
      Element* e = &n[nCounter--];
      to3(e->id, id);
      copy3(e->next, b[pos].next);
      to3(b[pos].next, nCounter + 1);
    }
  }

  void deleteMin() {
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
      //cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    if (from3(b[curPos].next) != MAX) {
      copy3(b[curPos].id, n[from3(b[curPos].next)].id);
      copy3(b[curPos].next, n[from3(b[curPos].next)].next);
    } else {
      to3(b[curPos].id, MAX);
    }
  }

  const id_slot & getMin() {
    while (from3(b[curPos].id) == MAX) {
      curPos = (curPos + 1) & mask;
      curKey++;
    }
    cout << curPos << endl;
    cout << from3(b[curPos].id) << endl;
    return from3(b[curPos].id);
  }

  const key_slot & getMinKey() const{
    return curKey;
  }

  const key_slot & getCurrentKey( const id_slot& id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  bool isReached( const id_slot& id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != numeric_limits<key_slot>::max()) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != numeric_limits<unsigned int>::max();
  }

  bool contains( const id_slot& id ) const {
  }

  void decreaseKey( const id_slot& id, const key_slot& new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    register unsigned int pos = new_key & mask;
    register unsigned int oldPos = p[id] & mask;

    p[id] = new_key;
    register Element* e = NULL;

    if (from3(b[oldPos].id) == id) {
      if (from3(b[oldPos].next) == MAX) {
        to3(b[oldPos].id, MAX);
      } else {
        e = &n[from3(b[oldPos].next)];
        copy3(b[oldPos].id, e->id);
        copy3(b[oldPos].next, e->next);
      }
    } else {
      Element* f = &b[oldPos];

      cout << id << endl;
      while (from3(n[from3(f->next)].id) != id) {
        //cout << from3(n[from3(f->next)].id) << endl;
        f = &n[from3(f->next)];
      }

      e = &n[from3(f->next)];
      copy3(f->next, e->next);
    }

    if (from3(b[pos].id) != MAX) {
      if (e == NULL) {
        e = &n[nCounter--];
      }
      to3(e->id, id);
      copy3(e->next, b[pos].next);
      to3(b[pos].next, e - n);
    } else {
      to3(b[pos].id, id);
    }
  }

  void increaseKey( const id_slot& id, const key_slot& new_key ) {
  }

  void updateKey( const id_slot& id, const key_slot& new_key ) {
  }

  void clear() {
    nCounter = pSize - 1;
    curPos = 0;
    curKey = 0;

    memset(b, 0xFF, maxWeight * sizeof(Element)); // bad style...
    //for (register unsigned int i = 0; i < maxWeight; i++) {
    //  b[i].next = NULL;
    //  b[i].key = UINT_MAX;
    //  //b[i].next = &b[i];
    //}

    memset(p, 0xFF, pSize * sizeof(key_slot));
    //fill_n(p, pSize, UINT_MAX);
  }
};


#endif /* BINARY_HEAP_HPP_ */
