/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>

#define DIRECT_IDS 4

using namespace std;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    Element* next;
  };

  struct Bucket {
    id_slot directIds[DIRECT_IDS];
    Element* indirectIds;
  };

  const unsigned int pSize;
  unsigned int maxWeight;
  unsigned int mask;
  unsigned int nCounter;
  unsigned int curPos;
  unsigned int curKey;
  bool isBig;

  Bucket* b;
  Element* n;
  key_slot* p;

  static void insertAfter(register Element* e, register Element* n) {
    n->next = e->next;
    e->next = n;
  }

  static void cutOutAfter(register Element* e) {
    e->next = e->next->next;
  }

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(0)
  {
    cout << "go" << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    p = new key_slot[pSize];
    n = new Element[pSize];
    b = new Bucket[maxWeight];
    //p = (key_slot*) (((unsigned long) (new char[(pSize)*sizeof(key_slot) + 256]) + 256) & ~0xFF);
    //n = (Element*) (((unsigned long) (new char[(pSize)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //b = (Element*) (((unsigned long) (new char[(maxWeight)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //aligned_storage<262144*sizeof(Element), alignment_of<Element>::value > buf;
    //b = reinterpret_cast<Element&>(buf);
  }

  inline size_t size() const{
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    //pushCounter++;

    //if (pushCounter == 1000) {
    //  pushCounter = 0;

    //  for (unsigned int i = 0; i < maxWeight; i++) {
    //    unsigned int count = 0;
    //    if (b[i].key != UINT_MAX) {
    //      count++;

    //      Element* e = b[i].next;
    //      while (e != &b[i]) {
    //        e = e->next;
    //        count++;
    //      }
    //    }
    //    cout << count << ",";
    //  }
    //  cout << endl;
    //}

    p[id] = key;
    if (key > maxWeight) {
      isBig = true;
    }

    unsigned int pos = key;
    if (isBig) {
      pos &= mask;
    }

    for (unsigned int i = 0; i < DIRECT_IDS; i++) {
      if (b[pos].directIds[i] == UINT_MAX) {
        //cout << "a" << endl;
        b[pos].directIds[i] = id;
        return;
      }
    }

    Element* e = &n[nCounter++];
    e->id = id;
    e->next = b[pos].indirectIds;
    b[pos].indirectIds = e;

    //if (b[pos].id != UINT_MAX) {
    //  Element* e = &n[nCounter++];
    //  e->id = id;
    //  insertAfter(&b[pos], e);
    //} else {
    //  b[pos].id = id;
    //  b[pos].next = &b[pos];
    //}
  }

  void deleteMin(){
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
      //cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    Bucket* bp = &(b[curPos]);

    if (bp->indirectIds != NULL) {
      //cout << bp->indirectIds << endl;
      bp->directIds[0] = bp->indirectIds->id;
      bp->indirectIds = bp->indirectIds->next;
      return;
    }

    unsigned int i = 0;
    while (i < DIRECT_IDS - 1) {
      //cout << "move" << endl;
      bp->directIds[i] = bp->directIds[i+1];

      i++;

      if (bp->directIds[i] == UINT_MAX) {
        return;
      }
    }

    bp->directIds[i] = UINT_MAX;

    //if (b[curPos].next == &b[curPos]) {
    //  b[curPos].id = UINT_MAX;
    //} else {
    //  cutOutAfter(&b[curPos]);
    //}
  }

  inline const id_slot & getMin() {
    while (b[curPos].directIds[0] == UINT_MAX) {
      curPos++;
      curKey++;
      if (isBig) {
        curPos &= mask;
      }
    }

    return b[curPos].directIds[0];
  }

  inline const key_slot & getMinKey() const{
    return curKey;
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != UINT_MAX) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != UINT_MAX;
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    //register unsigned int pos = new_key;
    //register unsigned int oldPos = p[id];
    //if (isBig) {
    //  pos &= mask;
    //  oldPos &= mask;
    //}
    //p[id] = new_key;
    //register Element* e = NULL;

    //if (b[oldPos].id == id) {
    //  if (b[oldPos].next == &b[oldPos]) {
    //    b[oldPos].id = UINT_MAX;
    //  } else {
    //    e = b[oldPos].next;
    //    b[oldPos].id = e->id;
    //    b[oldPos].next = e->next;
    //  }
    //} else {
    //  Element* f = &b[oldPos];

    //  while (f->next->id != id) {
    //    f = f->next;
    //  }

    //  e = f->next;

    //  cutOutAfter(f);
    //}

    //if (b[pos].id != UINT_MAX) {
    //  if (e == NULL) {
    //    e = &n[nCounter++];
    //  }
    //  e->id = id;
    //  insertAfter(&b[pos], e);
    //} else {
    //  b[pos].id = id;
    //  b[pos].next = &b[pos];
    //}
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    nCounter = 0;
    curPos = 0;
    curKey = 0;
    isBig = false;

    memset(b, 0xFF, maxWeight * sizeof(Bucket)); // bad style...
    for (register unsigned int i = 0; i < maxWeight; i++) {
      b[i].indirectIds = NULL;
    //  for (register unsigned int j = 0; j < DIRECT_IDS; j++) {
    //    b[i].directIds[j] = UINT_MAX;
    //  }
    //  //b[i].next = &b[i];
    }

    memset(p, 0xFF, pSize * sizeof(key_slot));
    //fill_n(p, pSize, UINT_MAX);
  }
};


#endif /* BINARY_HEAP_HPP_ */
