/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    key_slot key;
  };

  struct PosKey {
    unsigned int pos;
    key_slot key;
  };

  unsigned int _size;
  unsigned int finalLayerSize;
  unsigned int finalLayerDist;

  Element* heap;
  PosKey* pk;

  unsigned int maxSize;

  static unsigned int getParent(unsigned int& x)
  {
    return x >> 1;
  }

  static unsigned int getChild1(unsigned int& x)
  {
    return x << 1;
  }

  static unsigned int getChild2(unsigned int& x)
  {
    return (x << 1) + 1;
  }

public:
  PriorityQueue( const GraphType & graph )
  : maxSize(graph.numberOfNodes() + 2)
  {
    cout << "go" << endl;
    pk = new PosKey[maxSize];
    heap = new Element[maxSize];
    heap[0].key = meta_key_slot::min();
  }

  inline size_t size() const{
    return _size;
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/
    _size++;

    unsigned int hole = _size;
    unsigned int layerSize = finalLayerSize;
    unsigned int layerDist = finalLayerDist;
    if (finalLayerDist == 0) {
      finalLayerSize <<= 2;
      finalLayerDist = finalLayerSize - 1;
    } else {
      finalLayerDist--;
    }

    unsigned int pred = hole + layerDist - layerSize;
    layerSize >>= 2;
    layerDist >>= 2;
    pred = pred - layerDist;
    key_slot predKey = heap[pred].key;

    // terminates as there's a sentinel at 0
    while (predKey > key) {
      heap[hole].key = predKey;
      heap[hole].id = heap[pred].id;
      pk[heap[hole].id].pos = hole;
      hole = pred;
      pred = hole + layerDist - layerSize;
      layerSize >>= 2;
      layerDist >>= 2;
      pred = pred - layerDist;
      predKey = heap[pred].key;
    }

    heap[hole].key = key;
    heap[hole].id = id;

    pk[id].key = key;
    pk[id].pos = hole;
  }

  void deleteMin(){
#ifdef DEBUG/*{{{*/
      cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    unsigned int hole = 1;
    unsigned int succ = 2;
    unsigned int layerSize = 4;
    unsigned int layerPos = 0;
    unsigned int delta;
    key_slot compKey;
    key_slot minKey;

    finalLayerDist++;
    if (finalLayerDist == finalLayerSize) {
      finalLayerSize >>= 2;
      finalLayerDist = 0;
    }

    while (succ < _size) {
      minKey = heap[succ].key;
      delta = 0;

      compKey = heap[succ + 1].key;
      if (compKey < minKey) {
        minKey = compKey;
        delta = 1;
      }
      compKey = heap[succ + 2].key;
      if (compKey < minKey) {
        minKey = compKey;
        delta = 2;
      }
      compKey = heap[succ + 3].key;
      if (compKey < minKey) {
        minKey = compKey;
        delta = 3;
      }

      succ += delta;
      layerPos += delta;

      heap[hole].id = heap[succ].id;
      heap[hole].key = minKey;
      pk[heap[hole].id].pos = hole;

      hole = succ;
      succ = succ - layerPos + layerSize;
      layerPos <<= 2;
      succ += layerPos;
      layerSize <<= 2;
    }

    key_slot bubble = heap[_size].key;
    layerSize >>= 2;
    layerPos >>= 2;

    unsigned int layerDist = layerSize - layerPos - 1;
    unsigned int pred = hole + layerDist - layerSize;
    layerSize >>= 2;
    layerDist >>= 2;
    pred = pred - layerDist;

    while (heap[pred].key > bubble) {
      heap[hole] = heap[pred];
      pk[heap[hole].id].pos = hole;
      hole = pred;
      pred = hole + layerDist - layerSize;
      layerSize >>= 2;
      layerDist >>= 2;
      pred = pred - layerDist;
    }

    heap[hole].key = bubble;
    heap[hole].id = heap[_size].id;
    pk[heap[hole].id].pos = hole;

    heap[_size].key = meta_key_slot::max();
    _size--;
  }

  inline const id_slot & getMin() {
    return heap[1].id;
  }

  inline const key_slot & getMinKey() const{
    return heap[1].key;
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id]->key << endl;
#endif/*}}}*/
    return pk[id].key;
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != NULL) ? "1" : "0") << endl;
#endif/*}}}*/
    return pk[id].pos != 0;
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    //unsigned int hole = pk[id].pos;
    //unsigned int pred = hole >> 1;
    //key_slot predKey = heap[pred].key;

    //while (predKey > new_key) {
    //  heap[hole].key = predKey;
    //  heap[hole].id = heap[pred].id;
    //  pk[heap[hole].id].pos = hole;
    //  hole = pred;
    //  pred >>= 1;
    //  predKey = heap[pred].key;
    //}

    //heap[hole].key = new_key;
    //heap[hole].id = id;

    //pk[id].key = new_key;
    //pk[id].pos = hole;

    //unsigned int layerSize = 4 << (hole / 4);
    //unsigned int layerPos = 
    //unsigned int layerDist = layerSize - layerPos - 1;

    //unsigned int pred = hole + layerDist - layerSize;
    //layerSize >>= 2;
    //layerDist >>= 2;
    //pred = pred - layerDist;
    //key_slot predKey = heap[pred].key;

    //// terminates as there's a sentinel at 0
    //while (predKey > key) {
    //  heap[hole].key = predKey;
    //  heap[hole].id = heap[pred].id;
    //  pk[heap[hole].id].pos = hole;
    //  hole = pred;
    //  pred = hole + layerDist - layerSize;
    //  layerSize >>= 2;
    //  layerDist >>= 2;
    //  pred = pred - layerDist;
    //  predKey = heap[pred].key;
    //}
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    _size = 0;
    finalLayerSize = 1;
    finalLayerDist = 0;

    for (unsigned int i = 1; i < maxSize; i++) {
      // TODO: less
      heap[i].key = meta_key_slot::max();
    }

    for (unsigned int i = 0; i < maxSize; i++) {
      pk[i].pos = 0;
    }
    //memset(keys, meta_key_slot::max(), maxSize * sizeof(unsigned int));
  }
};


#endif /* BINARY_HEAP_HPP_ */
