/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>
#include <limits>

using namespace std;

#define TREESIZE 8388608
#define LOGTS 23

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    key_slot key;
  };

  const unsigned int treeSize;
  const unsigned int elementStartPos;
  unsigned int maxWeight;

  // Loser tree with winner in tree[0]
  Element* tree;

  bool isRight(unsigned int id, unsigned int level) {
    return id & (1 << (TS - 1 - level));
  }

public:
  PriorityQueue( const GraphType & graph )
  : treeSize(TREESIZE*2)
  //: treeSize(TREESIZE+graph.numberOfNodes())
  , elementStartPos(TREESIZE)
  , maxWeight(0)
  {
    cout << "go" << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    tree = new Element[treeSize];
  }

  size_t size() const{
  }

  bool empty() const {
#ifdef DEBUG/*{{{*/
    //cout << "size: " << size() << endl;
#endif/*}}}*/
    // TODO: Implement efficiently
    return false;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/
    unsigned int elemPos = elementStartPos + id;

    tree[elemPos].id = id; // TODO: optimize
    tree[elemPos].key = key;

    id_slot overId;
    key_slot overKey;

    if (key < tree[0].key) {
      overId = tree[0].id;
      overKey = tree[0].key;
      tree[0].id = id;
      tree[0].key = key;
    } else {
      overId = id;
      overKey = key;
      //winnerId = tree[0].id;
      //winnerKey = tree[0].key;
    }

    unsigned int level = 0;
    unsigned int pos = 1;

    while (pos < TREESIZE) {
      if (overKey > tree[pos].key) {
        overId = tree[pos].id;
        overKey = tree[pos].key;
        tree[pos].id = overId;
        tree[pos].key = overKey;
      }

      pos <<= 2;

      if (isRight(id, level++)) {
        pos++;
      }
    }
  }

  void deleteMin() {
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
#endif/*}}}*/
    tree[tree[0].id].key = UINT_MAX;

    redownTree();
    //reuppTree(tree[0]); // TODO: really?
  }

  const id_slot & getMin() const {
    return tree[0].id;
  }

  const key_slot & getMinKey() const{
    return tree[0].key;
  }

  const key_slot & getCurrentKey( const id_slot& id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return tree[elementStartPos + id].key;
  }

  bool isReached( const id_slot& id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != numeric_limits<key_slot>::max()) ? "1" : "0") << endl;
#endif/*}}}*/
    return tree[elementStartPos + id].key != numeric_limits<key_slot>::max();
  }

  bool contains( const id_slot& id ) const {
  }

  void decreaseKey( const id_slot& id, const key_slot& new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/
  }

  void increaseKey( const id_slot& id, const key_slot& new_key ) {
  }

  void updateKey( const id_slot& id, const key_slot& new_key ) {
  }

  void clear() {
    memset(tree, 0xFF, treeSize * sizeof(Element));
  }
};


#endif /* BINARY_HEAP_HPP_ */
