/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  class Element {
    public:
      id_slot id;
      key_slot key;
      Element* next;
      Element* prev;

      Element(id_slot i = UINT_MAX, key_slot k = UINT_MAX)
      : id(i)
      , key(k)
      , next(NULL)
      , prev(NULL)
      {}
  };

  unsigned int removedKey;

  unsigned char* bq;
  Element* keys; // index is id
  Element** ids; // index is key

  unsigned int pSize;
  unsigned int nCounter;
  unsigned int maxWeight;
  unsigned int mask;
  unsigned int curPos;

  static unsigned char msd(unsigned char c) {
    static const unsigned char LogTable256[256] =
    {
    #define LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
        0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
        LT(4), LT(5), LT(5), LT(6), LT(6), LT(6), LT(6),
        LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7)
    };

    return LogTable256[c];
  }

  static void insertBefore(register Element* e, register Element* n) {
    register Element* nn = e->prev;
#ifdef DEBUG
    cout << "nn: " << nn << endl;
#endif
    n->next = e;
    n->prev = nn;
    nn->next = n;
    e->prev = n;
  }

  static void insertAfter(register Element* e, register Element* n) {
    register Element* nn = e->next;
    n->prev = e;
    n->next = nn;
    nn->prev = n;
    e->next = n;
  }

  static void cutOut(Element* e) {
    register Element* p = e->prev;
    register Element* n = e->next;
    n->prev = p;
    p->next = n;
  }

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(262144)
  , mask(maxWeight - 1)
  {
    cout << "go" << endl;

    //for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
    //  unsigned int newWeight = graph.getEdge(eid).getWeight();
    //  //unsigned int newWeight = 2000;

    //  if (newWeight > maxWeight) {
    //    maxWeight = newWeight;
    //  }
    //}

    //maxWeight += 2;
    //cout << "maxWeight: " << maxWeight << endl;

    bq = new unsigned char[maxWeight >> 3];
    keys = new Element[pSize];
    ids = new Element*[maxWeight];

    for (int i = 0; i <= pSize; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      keys[i].id = i;
      keys[i].key = UINT_MAX;
    }

    //for (int i = 0; i < pSize; i++) {
    //  p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);
    //}
  }

  inline size_t size() const{
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  void printList(Element* e) {
    Element *x = e->next;
    cout << "----------" << endl;
    cout << e << "(id:" << e->id << ",key:" << e->key << "," << e->prev << "," << e->next << endl;
    while (x != e) {
      cout << x << "(id:" << x->id << ",key:" << x->key << "," << x->prev << "," << x->next << endl;
      x = x->next;
    }
    cout << "----------" << endl;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    unsigned int pos = key & mask;
    unsigned int bqPos = pos >> 3;
    unsigned int digit = pos & 7;

    //cout << "  pos: " << bqPos << ", digit: " << digit << endl;

    bq[bqPos] |= (1 << digit);
    //cout << digit << endl;
    keys[id].key = key;

    if (ids[key] == NULL) {
      ids[key] = &keys[id];
    } else {
      insertAfter(ids[key], &keys[id]);
    }
  }

  void deleteMin(){
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
#endif/*}}}*/
  }

  inline const id_slot & getMin() {
    while (bq[curPos] == 0) {
      curPos = (curPos + 1) & mask;
    }

    //cout << curPos << endl;
    //cout << (unsigned short) bq[curPos] << endl;
    unsigned char digit = msd(bq[curPos]);

    //cout << "  pos: " << curPos << ", digit: " << (unsigned int) digit << endl;

    unsigned int pos = (curPos << 3) ^ digit;
    //cout << "pos: " << pos << endl;
    Element* e = ids[pos];

    if (e->next == e) {
      //cout << "  before: " << bq[curPos] << endl;
      bq[curPos] &= (1 << digit) - 1;
      //cout << "  after:  " << bq[curPos] << endl;
      ids[pos] = NULL;
    } else {
      cutOut(e);

      if (ids[pos] == e) {
        ids[pos] = e->next;
      }
    }

    removedKey = e->key;

    return e->id;
  }

  inline const key_slot & getMinKey() const{
    return removedKey;
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << keys[id].key << endl;
#endif/*}}}*/
    return keys[id].key;
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((keys[id].key != UINT_MAX) ? "1" : "0") << endl;
#endif/*}}}*/
    return keys[id].key != UINT_MAX;
    //return p[id].key != UINT_MAX;
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    //unsigned int pos = new_key & mask;
    //unsigned int bqPos = pos / 8;
    //unsigned int digit = pos % 8;

    //Element* e = &keys[id];
    //unsigned int old_key = e->key;

    //cutOut(e);

    //if (ids[old_key] == e) { // only element
    //  ids[old_key] = NULL;
    //  unsigned int oldKeyPos = old_key & mask;
    //  unsigned int oldKeyBqPos = oldKeyPos >> 3;
    //  unsigned int oldKeyDigit = oldKeyPos & 7;
    //  bq[oldKeyBqPos] &= ~(1 << oldKeyDigit);
    //}

    //bq[bqPos] |= (1 << digit);
    //e->key = new_key;

    //if (ids[new_key] == NULL) {
    //  ids[new_key] = e;
    //} else {
    //  insertAfter(ids[new_key], e);
    //}
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    nCounter = 0;
    curPos = 0;

    for (register unsigned int i = 0; i <= pSize; i++) {
      keys[i].next = &keys[i];
      keys[i].prev = &keys[i];
    }

    memset(bq, 0, (maxWeight >> 3) * sizeof(unsigned char));
    memset(ids, 0, maxWeight * sizeof(Element*));

    //for (int i = 0; i < pSize; i++) {
    //  //n[i].key = UINT_MAX;
    //  p[i] = NULL;

    //  //if (p[i] != NULL) {
    //  //  delete p[i];
    //  //}
    //  //p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);

    //  //p[i]->id = i;
    //  //p[i]->next = NULL;
    //  //p[i]->prev = NULL;
    //}
  }
};


#endif /* BINARY_HEAP_HPP_ */
