/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

#include <vector>
#include <tr1/tuple>
#include <iostream>
#include <string>
#include <map>
#include <queue>
#include <tr1/unordered_map>

void debug( std::string s )
{
  std::cerr << s << std::endl;
}

//using namespace std;
//using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  typedef std::pair<id_slot,key_slot> id_key;
  typedef unsigned long size_type;

  class node
  {
  public:
    node(id_key const& d)
    : m_prev(0)
    , m_children(0)
    , m_data(d)
    { }

    static void insert(node** l, node* n) {
      n->m_next = *l;
      if (n->m_next != 0)
        n->m_next->m_prev = &n->m_next;
      n->m_prev = l;
      *l = n;
    }
    static void set_root(node** l, node* n) {
      n->m_next = 0;
      n->m_prev = l;
      *l = n;
    }
    static void remove(node* n) {
      debug("b.1.1.1");
      if (n->m_next != 0) {
        debug("b.1.1.2");
        n->m_next->m_prev = n->m_prev;
        debug("b.1.1.3");
      }
      debug("b.1.1.4");
      *n->m_prev = n->m_next;
      debug("b.1.1.5");
      n->m_prev = 0;
      debug("b.1.1.6");
    }

    id_key const& data() const     { return m_data; }
    void data(key_slot const& k) { m_data.second = k; }

    node** children() { return &m_children; }
    node*  begin()    { return m_children; }
    node*  next()     { return m_next; }
    node** prev() { return m_prev; }

  private:
    node** m_prev;
    node*  m_next;
    node*  m_children;
    id_key m_data;
  };

  typedef node*         pointer;

  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy
  //typedef storage_slot Storage;

  node* link(node* n1, node* n2) {
    if (n2->data().second > n1->data().second)
      {
        node::insert(n1->children(), n2);
        return n1;
      }
    else
      {
        node::insert(n2->children(), n1);
        return n2;
      }
  }

  node* recursive_link(node* n1, node* n2, node* c) {
    node::remove(n1);
    node::remove(n2);

    n1 = link(n1, n2);

    if (c)
      {
        if (c->next() != 0)
          c = recursive_link(c, c->next(), c->next()->next());
        else
          node::remove(c);
        n1 = link(n1, c);
      }

    return n1;
  }

  void  unlink(node* n) {
    debug("b.1.1");
    node::remove(n);
    debug("b.1.2");

    if (node* c = n->begin())
      {
        debug("b.1.3");
        if (c->next() != 0) {
          debug("b.1.4");
          c = recursive_link(c, c->next(), c->next()->next());
          debug("b.1.5");
        } else {
          debug("b.1.6");
          node::remove(c);
          debug("b.1.7");
        }

        node::set_root(&m_root, m_root == 0? c: link(m_root, c));
        debug("b.1.8");
      }
  }

  size_type m_size;
  node*     m_root;
  std::tr1::unordered_map<id_slot,pointer> m_pointers;
  //std::tr1::unordered_map<id_slot,bool> m_reached;

public:
  class iterator
  {
    friend class PriorityQueue;
    explicit iterator(node* root) { if (root != 0) m_queue.push(root); }

  public:
    iterator() {}

    id_key const& operator* () const { return m_queue.front()->data(); }
    id_key const* operator->() const { return &m_queue.front()->data(); }

    iterator& operator++()
      {
        node* tmp = m_queue.front()->begin();
        m_queue.pop();
        for (; tmp != 0; tmp = tmp->next())
          m_queue.push(tmp);
        return *this;
      }
    iterator  operator++(int)
      {
        iterator rc(*this);
        operator++();
        return rc;
      }

    bool operator== (iterator const& it) const
      {
        return m_queue.empty() == it.m_queue.empty()
               && (m_queue.empty() || &m_queue.front() == &it.m_queue.front());
      }
    bool operator!= (iterator const& it) const { return !operator== (it); }

  private:
    std::queue<node*> m_queue;
  };

  iterator  begin() const { return iterator(m_root); }
  iterator  end() const { return iterator(); }

  //typedef T             value_type;
  //typedef T&            reference;
  //typedef T const&      const_reference;
  //typedef iterator      const_iterator;

  PriorityQueue( const GraphType & graph )
  : m_size(0)
  , m_root(0)
  , m_pointers()
  //, m_reached()
  {
    debug("PriorityQueue()");
  }

  ~PriorityQueue()
  {
    iterator e = end();
    iterator it = begin();
    while (it != e)
      {
        node* del = it.m_queue.front();
        ++it;
        delete del;
      }
  }

  void remove(pointer n)
  {
    unlink(n);

    --m_size;
    delete n;
  }

  void change(pointer n, key_slot k)
  {
    debug("b.1");
    unlink(n);

    debug("b.2");
    n->data(k);
    debug("b.3");
    node::set_root(&m_root, m_root == 0? n: link(n, m_root));
    debug("b.4");
  }


  size_t size() const{
    debug("size()");
    return m_size;
  }

  bool empty() const {
    debug("empty()");
    return m_size == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
    debug("push()");
    node* rc = new node(std::make_pair(id,key));
    node::set_root(&m_root, m_size++ == 0? rc: link(rc, m_root));

    m_pointers[id] = rc;
    //m_reached[id] = true;
  }

  void deleteMin(){
    debug("deleteMin()");
    remove(m_root);
  }

  const id_slot & getMin() const{
    debug("getMin()");
    return m_root->data().first;
  }

  const key_slot & getMinKey() const {
    debug("getMinKey()");
    return m_root->data().second;
  }

  const key_slot & getCurrentKey( id_slot id ) {
    debug("getCurrentKey()");

    return m_pointers[id]->data().second;
  }

  bool isReached( id_slot id ) const {
    debug("isReached()");
    return m_pointers.count(id) != 0;
  }

  bool contains( id_slot id ) const {
    debug("contains()");
    return m_pointers.count(id) != 0;
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
    debug("decreaseKey()");

    debug("a");
    pointer x = m_pointers[id];
    debug("b");
    change(x, new_key);
    debug("c");
  }

  void increaseKey( id_slot id, key_slot new_key ){
    debug("increaseKey()");

    change(m_pointers[id], new_key);
  }

  void updateKey( id_slot id, key_slot new_key ){
    debug("updateKey()");

    change(m_pointers[id], new_key);
  }

  void clear(){
    debug("clear()");
  }
};


#endif /* BINARY_HEAP_HPP_ */
