/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

using namespace std;
using namespace std::tr1;

#define C 16
#define R 1

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename id_slot, typename key_slot, typename meta_key_slot >
class RadixHeap{
private:
  RadixHeap( const RadixHeap & ){}  //do not copy
  void operator=( const RadixHeap& ){}  //really, do not copy

  // Attributes/*{{{*/
  class Element {
    public:
      id_slot id;
      key_slot key;
      Element* next;
      Element* prev;

      Element(id_slot i = UINT_MAX, key_slot k = UINT_MAX)
      : id(i)
      , key(k)
      , next(NULL)
      , prev(NULL)
      {}
  };

  unsigned int _size;
  Element* b;
  Element** p;
  Element* n;
  unsigned int pSize;
  unsigned int nCounter;
  Element* f;
/*}}}*/
  static unsigned int msd(unsigned int a, unsigned int b) {/*{{{*/
#ifdef DEBUG
    cout << "msd(" << a << "," << b << ")" << endl;
#endif
    register unsigned int n = a^b;

    static const char LogTable256[256] =
    {
    #define LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
        0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
        LT(5), LT(6), LT(6), LT(7), LT(7), LT(7), LT(7),
        LT(8), LT(8), LT(8), LT(8), LT(8), LT(8), LT(8), LT(8)
    };

    if (n & 0xFFFF0000) {
      return C;
    }

    register unsigned int t = n >> 8;
    if (!t) {
      return LogTable256[n];
    }
    return 8 + LogTable256[t];
  }
/*}}}*/
  // insert n before e/*{{{*/
  static void insertBefore(register Element* e, register Element* n) {
    register Element* nn = e->prev;
#ifdef DEBUG
    cout << "nn: " << nn << endl;
#endif
    n->next = e;
    n->prev = nn;
    nn->next = n;
    e->prev = n;
  }
/*}}}*/
  static void insertAfter(register Element* e, register Element* n) {/*{{{*/
    register Element* nn = e->next;
    n->prev = e;
    n->next = nn;
    nn->prev = n;
    e->next = n;
  }
/*}}}*/
  static void cutOut(Element* e) {/*{{{*/
    register Element* p = e->prev;
    register Element* n = e->next;
    n->prev = p;
    p->next = n;
    // set e->next = NULL, e->prev = NULL?
    //e->next = NULL;
    //e->prev = NULL;
  }
/*}}}*/
  inline void insert(unsigned int pos, Element* n) {/*{{{*/
    // only first element in b[pos] has to be minimal
    register Element* bp = &b[pos];
    register Element* bpn = bp->next;
    if (n->key < bpn->key) {
      insertAfter(bp, n); // new minimum
    } else {
      insertAfter(bpn, n); // don't care
      //insertBefore(bp, n); // don't care
    }
  }
/*}}}*/
public:
  RadixHeap()/*{{{*/
  {
  }
/*}}}*/
  RadixHeap( unsigned int nNodes )/*{{{*/
  : pSize(nNodes)
  {
    cout << "go" << endl;
#ifdef DEBUG
    cout << "n: " << nNodes << endl;
#endif
    b = new Element[C+1];
    p = new Element *[nNodes];
    n = new Element[nNodes];
    f = new Element(UINT_MAX, UINT_MAX);

    for (int i = 0; i <= C; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      b[i].id = UINT_MAX;
      b[i].key = UINT_MAX;
#ifdef DEBUG
      cout << "b[" << i << "] = " << &b[i] << endl;
#endif
    }

    //for (int i = 0; i < pSize; i++) {
    //  p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);
    //}
  }
/*}}}*/
  inline size_t size() const{/*{{{*/
    return _size;
  }
/*}}}*/
  inline bool empty() const {/*{{{*/
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    //return false;
    //return b[0].next == &b[0]; ?
    return size() == 0;
  }
/*}}}*/
  void printList(Element* e) {/*{{{*/
    Element *x = e->next;
    cout << "----------" << endl;
    cout << e << "(id:" << e->id << ",key:" << e->key << "," << e->prev << "," << e->next << endl;
    while (x != e) {
      cout << x << "(id:" << x->id << ",key:" << x->key << "," << x->prev << "," << x->next << endl;
      x = x->next;
    }
    cout << "----------" << endl;
  }
/*}}}*/
  void push( const id_slot & id, const key_slot & key ){/*{{{*/
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

#ifdef DEBUG
    printList(f);
#endif
    register Element* e = &n[nCounter++];
    p[id] = e;
    e->id = id;
    e->key = key;
    ++_size;

    // Register lohnen hier nicht
    //cout << e << endl;
    if (b[0].next == &b[0]) {
#ifdef DEBUG/*{{{*/
      cout << "  on b[0]" << endl;
#endif/*}}}*/
      insertAfter(&b[0], e);
    } else if (key < b[0].next->key) {
#ifdef DEBUG/*{{{*/
      cout << "  on f" << endl;
#endif/*}}}*/
      // Maybe insert in f at right place? arbitrary according to sanders
      // either sanders is wrong or i misunderstood, has to be sorted

      register Element* n = f->next;
      while (key > n->key) {
        n = n->next;
      }
#ifdef DEBUG/*{{{*/
      cout << "n: " << n << ", e: " << e << endl;
      cout << "n->n: " << n->next << ", n->p: " << n->prev << endl;
      cout << "e->n: " << e->next << ", e->p: " << e->prev << endl;
#endif/*}}}*/
      insertBefore(n, e);

      //insertAfter(f, e);
    } else {
#ifdef DEBUG/*{{{*/
      cout << "  on " << msd(b[0].next->key, key) << endl;
#endif/*}}}*/
      insert(msd(b[0].next->key, key), e);
    }
  }
/*}}}*/
  void deleteMin(){/*{{{*/
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
#endif/*}}}*/
    --_size;
#ifdef DEBUG
    printList(f);
#endif

    if (f->next != f) {
#ifdef DEBUG/*{{{*/
      cout << "  from f: " << f->next->id << "," << f->next->key << endl;
#endif/*}}}*/
      cutOut(f->next);
#ifdef DEBUG
      cout << "here: " << f->next << "," << f << endl;
#endif
      return;
    }

#ifdef DEBUG/*{{{*/
    cout << "  from b[0]: " << b[0].next->id << "," << b[0].next->key << endl;
#endif/*}}}*/
    cutOut(b[0].next);

    if (&b[0] == b[0].next) {
      for (register unsigned int i = 1; i <= C; i++) {
        if (&b[i] != b[i].next) {
          register Element* e = b[i].next->next;
          register unsigned int b0nk = b[i].next->key;
          insertAfter(&b[0], b[i].next);
#ifdef DEBUG/*{{{*/
          cout << "  new b[0]: " << b[0].next->id << "," << b[0].next->key << endl;
#endif/*}}}*/
          //cout << "b[i]->id: " << b[i]->id << endl;
          b[i].next = &b[i];
          b[i].prev = &b[i];
          //int u = 0;
          while (e != &b[i]) {
            //u += 1;
#ifdef DEBUG/*{{{*/
            cout << "  new pos for " << e->id << "," << e->key << ": " << msd(b[0].next->key, e->key) << endl;
#endif/*}}}*/
            register Element* tmp = e->next;
            insert(msd(b0nk, e->key), e);
            //cout << "insert done" << endl;
            e = tmp;
          }
          //cout << "el: " << u << endl;
          return;
        }
      }
    }
  }
/*}}}*/
  inline const id_slot & getMin() const{/*{{{*/
    register Element* fn = f->next;
    if (fn == f)
      return b[0].next->id;
    return fn->id;
  }

  inline const key_slot & getMinKey() const{
    register Element* fn = f->next;
    if (fn == f)
      return b[0].next->key;
    return fn->key;
  }
/*}}}*/
  inline const key_slot & getCurrentKey( id_slot id ) {/*{{{*/
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id]->key << endl;
#endif/*}}}*/
    return p[id]->key;
  }
/*}}}*/
  inline bool isReached( id_slot id ) const {/*{{{*/
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != NULL) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != NULL;
    //return p[id].key != UINT_MAX;
  }

  inline bool contains( id_slot id ) const {
  }
/*}}}*/
  void decreaseKey( id_slot id, key_slot new_key ) {/*{{{*/
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
    printList(f);
#endif/*}}}*/

    register Element* e = p[id];
    register Element* b0n = b[0].next;
    register unsigned int b0nk = b0n->key;

    if (b0n != e) { // kein problem wenn zwei in b[0]
      cutOut(e);
      e->key = new_key;
      if (e->prev->id == UINT_MAX && e->next->id != UINT_MAX) { // removed the min from some bucket
        register Element* min = e->prev;
        for (register Element *x = e->next; x != e->prev; x = x->next) {
          if (x->key < min->key) {
            min = x;
          }
        }
        cutOut(min);
        insertAfter(e->prev, min);
      }
      //e->next = NULL; // not necc?

      if (new_key < b0nk) {
        //cout << "1" << endl;
#ifdef DEBUG/*{{{*/
        cout << "  into f" << endl;
#endif/*}}}*/
        register Element* n = f->next;
        register unsigned int ek = e->key;
        while (e->key > n->key) {
          //cout << n << endl;
          n = n->next;
        }
        insertBefore(n, e);

        //insertAfter(f, e);
        return;
      }

      //cout << "2" << endl;
#ifdef DEBUG/*{{{*/
      cout << "  into " << msd(b0nk, new_key) << endl;
#endif/*}}}*/
      insert(msd(b0nk, new_key), e);
      return;
    }

#ifdef DEBUG/*{{{*/
    cout << "  b[0] == NULL || decrease b[0]" << endl;
#endif/*}}}*/
    // decreasing b[0]
    e->key = new_key;
    cutOut(b[0].next);

    register Element* n = f->next;
    while (new_key > n->key) {
      n = n->next;
    }
    insertBefore(n, e);

    for (register unsigned int i = 1; i <= C; i++) {
      if (&b[i] != b[i].next) {
        register Element* e2 = b[i].next->next;
        register unsigned int b0nk = b[i].next->key;
        insertAfter(&b[0], b[i].next);
#ifdef DEBUG/*{{{*/
        cout << "  new b[0]: " << b[0].next->id << "," << b[0].next->key << endl;
#endif/*}}}*/
        //cout << "b[i]->id: " << b[i]->id << endl;
        b[i].next = &b[i];
        b[i].prev = &b[i];
        //int u = 0;
        while (e2 != &b[i]) {
          //u += 1;
#ifdef DEBUG/*{{{*/
          cout << "  new pos for " << e->id << "," << e->key << ": " << msd(b[0].next->key, e->key) << endl;
#endif/*}}}*/
          register Element* tmp = e2->next;
          insert(msd(b0nk, e2->key), e2);
          e2 = tmp;
        }
        //cout << "el: " << u << endl;
        return;
      }
    }
    // No non-empty b[i] found, TODO: write better
    Element *fp = f->prev;
    cutOut(fp);
    insertAfter(&b[0],fp);
  }
/*}}}*/
  inline void increaseKey( id_slot id, key_slot new_key ) {/*{{{*/
  }
/*}}}*/
  inline void updateKey( id_slot id, key_slot new_key ) {/*{{{*/
  }
/*}}}*/
  inline void clear() {/*{{{*/
    _size = 0;
    nCounter = 0;

    f->next = f;
    f->prev = f;

    for (register unsigned int i = 0; i <= C; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      b[i].next = &b[i];
      b[i].prev = &b[i];
    }

    memset(p, 0, pSize * sizeof(Element*));
    //for (int i = 0; i < pSize; i++) {
    //  //n[i].key = UINT_MAX;
    //  p[i] = NULL;

    //  //if (p[i] != NULL) {
    //  //  delete p[i];
    //  //}
    //  //p[i] = NULL;
    //  //p[i] = new Element(i, UINT_MAX);

    //  //p[i]->id = i;
    //  //p[i]->next = NULL;
    //  //p[i]->prev = NULL;
    //}
  }
/*}}}*/
};

template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue {
private:
  PriorityQueue( const PriorityQueue& ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  RadixHeap<id_slot, key_slot, meta_key_slot> *rh[R];
  unsigned int nNodes;
  unsigned int sizeRH;

public:
  PriorityQueue( const GraphType& graph )
  : nNodes(graph.numberOfNodes())
  , sizeRH(graph.numberOfNodes() / R)
  {
    for (unsigned int i = 0; i < R; i++) {
      rh[i] = new RadixHeap<id_slot, key_slot, meta_key_slot>(sizeRH);
    }
  }

  inline size_t size() const
  {
    unsigned int s = 0;

    for (unsigned int i = 0; i < R; i++) {
      s += rh[i]->size();
    }

    return s;
  }

  inline bool empty() const
  {
    return size() == 0;
  }

  void push(const id_slot& id, const key_slot& key)
  {
    unsigned int r = floor(id / (nNodes / R));

    rh[r]->push(id - r * sizeRH, key);
  }

  void deleteMin()
  {
    unsigned int min = UINT_MAX;
    unsigned int minIndex = 0;

    for (unsigned int i = 0; i < R; i++) {
      if (!rh[i]->empty() && rh[i]->getMinKey() < min) {
        min = rh[i]->getMinKey();
        minIndex = i;
      }
    }

    rh[minIndex]->deleteMin();
  }

  inline const id_slot getMin()
  {
    unsigned int min = UINT_MAX;
    unsigned int minIndex = 0;

    for (unsigned int i = 0; i < R; i++) {
      if (!rh[i]->empty() && rh[i]->getMinKey() < min) {
        min = rh[i]->getMinKey();
        minIndex = i;
      }
    }

    return rh[minIndex]->getMin() + minIndex * sizeRH;
  }

  inline const key_slot getMinKey()
  {
    unsigned int min = UINT_MAX;
    unsigned int minIndex = 0;

    for (unsigned int i = 0; i < R; i++) {
      if (!rh[i]->empty() && rh[i]->getMinKey() < min) {
        min = rh[i]->getMinKey();
        minIndex = i;
      }
    }

    return rh[minIndex]->getMinKey();
  }

  inline const key_slot& getCurrentKey(id_slot id)
  {
    unsigned int r = floor(id / (nNodes / R));

    return rh[r]->getCurrentKey(id - r * sizeRH);
  }

  inline bool isReached(id_slot id) const
  {
    unsigned int r = floor(id / (nNodes / R));

    return rh[r]->isReached(id - r * sizeRH);
  }

  void decreaseKey(id_slot id, key_slot new_key)
  {
    unsigned int r = floor(id / (nNodes / R));

    return rh[r]->decreaseKey(id - r * sizeRH, new_key);
  }

  void clear()
  {
    for (unsigned int i = 0; i < R; i++) {
      rh[i]->clear();
    }
  }
};

#endif /* BINARY_HEAP_HPP_ */
