/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#ifdef DEBUG
#define Debug(x) x
#else
#define Debug(x) 
#endif

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>
#include <limits>

using namespace std;

#define BS 4096

template <typename id_slot, typename key_slot>
struct IKElement {
  id_slot id;
  key_slot key;
};

template <typename id_slot, typename key_slot, typename meta_key_slot, int capacity>
class BinaryHeap {
  private:/*{{{*/
    typedef IKElement<id_slot, key_slot> Element;
    Element data[capacity + 2];
    int size; // current size
/*}}}*/
  public:
    BinaryHeap()/*{{{*/
    {
    }
/*}}}*/
    void clear()/*{{{*/
    {
      size = 0;

      data[0].key = meta_key_slot::min();

      // TODO: memcpy?
      for (int i = 1; i <= capacity + 1; i++) {
        data[i].key = meta_key_slot::max();
      }
    }
/*}}}*/
    id_slot& getMin() /*{{{*/
    {
      Debug(cout << "BH::getMin() = " << data[1].id << endl);

      return data[1].id;
    }
/*}}}*/
    key_slot& getMinKey() /*{{{*/
    {
      Debug(cout << "BH::getMinKey() = " << data[1].key << endl);

      return data[1].key;
    }
/*}}}*/
    void deleteMin()/*{{{*/
    {
      Debug(cout << "BH::deleteMin()" << endl);

      int hole = 1;
      int succ = 2;

      // move up min-path
      while (succ < size) {
        key_slot key1 = data[succ].key;
        key_slot key2 = data[succ + 1].key;

        if (key1 > key2) {
          succ++;
          data[hole].key = key2;
          data[hole].id = data[succ].id;
        } else {
          data[hole].key = key1;
          data[hole].id = data[succ].id;
        }

        hole = succ;
        succ <<= 1;
      }

      // bubble up rightmost element
      key_slot bubble = data[size].key;
      int pred = hole >> 1;

      // must terminate since infimum at root
      while (data[pred].key > bubble) {
        data[hole] = data[pred];
        hole = pred;
        pred >>= 1;
      }

      // move data to hole
      data[hole].key = bubble;
      data[hole].id = data[size].id;

      data[size].key = meta_key_slot::max();
      size -= 1;
    }
/*}}}*/
    void insert(id_slot id, key_slot key)/*{{{*/
    {
      Debug(cout << "BH::insert(" << id << "," << key << ")" << endl);
      size++;
      int hole = size;
      int pred = hole >> 1;
      key_slot predKey = data[pred].key;

      // terminates due to sentinel at 0
      while (predKey > key) {
        data[hole].key = predKey;
        data[hole].id = data[pred].id;
        hole = pred;
        pred >>= 1;
        predKey = data[pred].key;
      }

      data[hole].key = key;
      data[hole].id = id;
    }
/*}}}*/
    // not to be called for elements not in binary heap
    void decreaseKey(id_slot id, key_slot newKey)/*{{{*/
    {
      Debug(cout << "BH::decreaseKey(" << id << "," << newKey << ")" << endl);

      int hole = 1;

      while (id != data[hole].id) {
        hole++;
      }

      int pred = hole >> 1;

      // must terminate since infimum at root
      while (data[pred].key > newKey) {
        data[hole] = data[pred];
        hole = pred;
        pred >>= 1;
      }

      // reinsert element
      data[hole].key = newKey;
      data[hole].id = id;
    }
/*}}}*/
    void sortOutAndEmpty(Element *out)/*{{{*/
    {
      Debug(cout << "BH::sortOutAndEmpty(" << out << ")" << endl);

      Element* const outEnd = out + size;
      Element* const root = data + 1;

      while (out < outEnd) {
        // copy minimum
        *out = *root;
        out++;

        // bubble up second smallest as in deleteMin
        int hole = 1;
        int succ = 2;

        while (succ <= size) { // TODO: Warum <=, nicht wie oben < ?
          key_slot key1 = data[succ].key;
          key_slot key2 = data[succ + 1].key;

          if (key1 > key2) {
            succ++;
            data[hole].key = key2;
            data[hole].id = data[succ].id;
          } else {
            data[hole].key = key1;
            data[hole].id = data[succ].id;
          }

          hole = succ;
          succ <<= 1;
        }

        // mark hole as deleted
        data[hole].key = meta_key_slot::max();
      }

      size = 0;
    }
/*}}}*/
    int getSize() const/*{{{*/
    {
      Debug(cout << "BH::getSize() = " << size << endl);

      return size;
    }
/*}}}*/
};

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    unsigned int next;
    //Element* next;
    // unsigned char id[3];
    // unsigned char next[3];
  };

  const unsigned int pSize;
  const unsigned int mask;
  unsigned int maxWeight;
  unsigned int nCounter;
  unsigned int curPos;
  unsigned int curKey;

  Element* b;
  Element* n;
  key_slot* p;
  BinaryHeap<id_slot, key_slot, meta_key_slot, 32768> bh;

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , mask(BS - 1)
  , maxWeight(0)
  , bh()
  {
    cout << "go" << endl;
    cout << sizeof(id_slot) << endl;
    cout << sizeof(Element*) << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    //mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    n = new Element[pSize];
    p = new key_slot[pSize];
    //b = new Element[maxWeight];
    b = new Element[BS];
    //p = (key_slot*) (((unsigned long) (new char[(pSize)*sizeof(key_slot) + 256]) + 256) & ~0xFF);
    //n = (Element*) (((unsigned long) (new char[(pSize)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //b = (Element*) (((unsigned long) (new char[(maxWeight)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //aligned_storage<262144*sizeof(Element), alignment_of<Element>::value > buf;
    //b = reinterpret_cast<Element&>(buf);
  }

  size_t size() const{
  }

  bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    p[id] = key;

    if (key > curKey + BS) {
      bh.insert(id, key);
      return;
    }

    unsigned int pos = key & mask;

    if (b[pos].id == numeric_limits<id_slot>::max()) {
      b[pos].id = id;
    } else {
      Element* e = &n[nCounter];
      e->id = id;
      e->next = b[pos].next;
      b[pos].next = nCounter--;
    }
  }

  void deleteMin() {
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
      //cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    if (bh.getMinKey() == curKey) {
      bh.deleteMin();
    } else if (b[curPos].next != numeric_limits<unsigned int>::max()) {
      b[curPos].id = n[b[curPos].next].id;
      b[curPos].next = n[b[curPos].next].next;
    } else {
      b[curPos].id = numeric_limits<id_slot>::max();
    }
  }

  const id_slot & getMin() {
    key_slot bhKey = bh.getMinKey();
    while (b[curPos].id == numeric_limits<id_slot>::max()) {
      curPos = (curPos + 1) & mask;
      curKey++;
      if (curKey == bhKey) {
        return bh.getMin();
      }
    }
    return b[curPos].id;
  }

  const key_slot & getMinKey() const{
    return curKey;
  }

  const key_slot & getCurrentKey( const id_slot& id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  bool isReached( const id_slot& id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != numeric_limits<key_slot>::max()) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != numeric_limits<key_slot>::max();
  }

  bool contains( const id_slot& id ) const {
  }

  void decreaseKey( const id_slot& id, const key_slot& new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    register unsigned int pos = new_key & mask;
    register unsigned int oldPos = p[id] & mask;

    p[id] = new_key;
    register Element* e = NULL;

    if (b[oldPos].id == id) {
      if (b[oldPos].next == numeric_limits<unsigned int>::max()) {
        b[oldPos].id = numeric_limits<id_slot>::max();
      } else {
        e = &n[b[oldPos].next];
        b[oldPos].id = e->id;
        b[oldPos].next = e->next;
      }
    } else {
      Element* f = &b[oldPos];

      if (f->next == numeric_limits<unsigned int>::max()) {
        //bh.decreaseKey(id, new_key);
        return;
      }

      while (n[f->next].id != id) {
        f = &n[f->next];

        if (f->next == numeric_limits<unsigned int>::max()) {
          //bh.decreaseKey(id, new_key);
          return;
        }
      }

      e = &n[f->next];
      f->next = e->next;
    }

    if (b[pos].id != numeric_limits<id_slot>::max()) {
      if (e == NULL) {
        e = &n[nCounter--];
      }
      e->id = id;
      e->next = b[pos].next;
      b[pos].next = e - n;
    } else {
      b[pos].id = id;
    }
  }

  void increaseKey( const id_slot& id, const key_slot& new_key ) {
  }

  void updateKey( const id_slot& id, const key_slot& new_key ) {
  }

  void clear() {
    nCounter = pSize - 1;
    curPos = 0;
    curKey = 0;

    bh.clear();

    memset(b, 0xFF, BS * sizeof(Element)); // bad style...
    //memset(b, 0xFF, maxWeight * sizeof(Element)); // bad style...
    //for (register unsigned int i = 0; i < maxWeight; i++) {
    //  b[i].next = NULL;
    //  b[i].key = UINT_MAX;
    //  //b[i].next = &b[i];
    //}

    memset(p, 0xFF, pSize * sizeof(key_slot));
    //fill_n(p, pSize, UINT_MAX);
  }
};


#endif /* BINARY_HEAP_HPP_ */
