/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#include <iostream>
#include <string>
#include <cstring>
#include <limits.h>
#include <limits>

using namespace std;

#define BS 65536
#define MASK 65535

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  struct Element {
    id_slot id;
    key_slot key;
    unsigned int next;
    //Element* next;
    // unsigned char id[3];
    // unsigned char next[3];
  };

  const unsigned int pSize;
  unsigned int maxWeight;
  // const unsigned int mask;
  unsigned int nCounter;
  unsigned int curPos;
  unsigned int curKey;

  Element* b;
  Element* n;
  key_slot* p;

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  //, mask(BS - 1)
  , maxWeight(0)
  {
    cout << "go" << endl;
    cout << sizeof(id_slot) << endl;
    cout << sizeof(Element*) << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    //mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    p = new key_slot[pSize];
    n = new Element[pSize];
    b = new Element[BS];
    //p = (key_slot*) (((unsigned long) (new char[(pSize)*sizeof(key_slot) + 256]) + 256) & ~0xFF);
    //n = (Element*) (((unsigned long) (new char[(pSize)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //b = (Element*) (((unsigned long) (new char[(maxWeight)*sizeof(Element) + 256]) + 256) & ~0xFF);
    //aligned_storage<262144*sizeof(Element), alignment_of<Element>::value > buf;
    //b = reinterpret_cast<Element&>(buf);
  }

  size_t size() const{
  }

  bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/

    p[id] = key;
    unsigned int pos = key & MASK;
    //cout << pos << endl;
    //cout << &b[pos] << endl;
    //cout << b[pos].id << endl;
    //cout << b[pos].key << endl;
    //cout << b[pos].next << endl;

    if (b[pos].key == numeric_limits<key_slot>::max()) {
      b[pos].id = id;
      b[pos].key = key;
    } else if (key < b[pos].key) {
      Element* e = &n[nCounter];
      e->id = b[pos].id;
      e->key = b[pos].key;
      e->next = b[pos].next;
      b[pos].id = id;
      b[pos].key = key;
      b[pos].next = nCounter--;
    } else {
      Element* e = &n[nCounter];
      e->id = id;
      e->key = key;
      if (key == b[pos].key) {
        //cout << "a" << endl;
        e->next = b[pos].next;
        b[pos].next = nCounter--;
      } else {
        //cout << "b" << endl;
        Element* f = &b[pos];
        while (f->next < pSize && n[f->next].key < key) {
          f = &n[f->next];
        }
        e->next = f->next;
        f->next = nCounter--;
      }
    }
  }

  void deleteMin() {
#ifdef DEBUG/*{{{*/
      cout << "deleteMin()" << endl;
      //cout << "deleteMin() = " << b[curPos].next << ", id: " << b[curPos].next->id << ", key: " << b[curPos].next->key << endl;
#endif/*}}}*/
    if (b[curPos].next != numeric_limits<unsigned int>::max()) {
      b[curPos].id = n[b[curPos].next].id;
      b[curPos].key = n[b[curPos].next].key;
      b[curPos].next = n[b[curPos].next].next;
    } else {
      b[curPos].key = numeric_limits<key_slot>::max();
      b[curPos].next = numeric_limits<unsigned int>::max();
    }
  }

  const id_slot & getMin() {
    while (b[curPos].key != curKey) {
      curPos = (curPos + 1) & MASK;
      curKey++;
    }
    return b[curPos].id;
  }

  const key_slot & getMinKey() const{
    return curKey;
  }

  const key_slot & getCurrentKey( const id_slot& id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  bool isReached( const id_slot& id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != numeric_limits<key_slot>::max()) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != numeric_limits<key_slot>::max();
  }

  bool contains( const id_slot& id ) const {
  }

  void decreaseKey( const id_slot& id, const key_slot& new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    register unsigned int pos = new_key & MASK;
    register unsigned int oldPos = p[id] & MASK;

    p[id] = new_key;
    register Element* e = NULL;

    if (b[oldPos].id == id) {
      if (b[oldPos].next == numeric_limits<unsigned int>::max()) {
        b[oldPos].key = numeric_limits<key_slot>::max();
      } else {
        e = &n[b[oldPos].next];
        b[oldPos].id = e->id;
        b[oldPos].key = e->key;
        b[oldPos].next = e->next;
      }
    } else {
      Element* f = &b[oldPos];

      while (n[f->next].id != id) {
        f = &n[f->next];
      }

      e = &n[f->next];
      f->next = e->next;
    }

    if (b[pos].key != numeric_limits<key_slot>::max()) {
      if (e == NULL) {
        e = &n[nCounter--];
      }
      e->id = id;
      e->key = new_key;
      e->next = b[pos].next;
      b[pos].next = e - n;
    } else {
      b[pos].id = id;
      b[pos].key = new_key;
    }
  }

  void increaseKey( const id_slot& id, const key_slot& new_key ) {
  }

  void updateKey( const id_slot& id, const key_slot& new_key ) {
  }

  void clear() {
    nCounter = pSize - 1;
    curPos = 0;
    curKey = 0;

    memset(b, 0xFF, BS * sizeof(Element)); // bad style...
    //for (register unsigned int i = 0; i < maxWeight; i++) {
    //  b[i].next = NULL;
    //  b[i].key = UINT_MAX;
    //  //b[i].next = &b[i];
    //}

    memset(p, 0xFF, pSize * sizeof(key_slot));
    //fill_n(p, pSize, UINT_MAX);
  }
};


#endif /* BINARY_HEAP_HPP_ */
