/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

#include <vector>
#include <iostream>
#include <string>
#include <tr1/unordered_map>
#include <assert.h>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  typedef pair<id_slot,key_slot> elementT;
  typedef vector<elementT> vectorT;
  typedef typename vectorT::iterator vectorTI;
  typedef typename vectorT::const_iterator vectorTCI;

  vectorT v;
  vectorTI min;
  unordered_map<id_slot,bool> reached;
  id_slot _null;

  void debug(string s) const {
    //cout << s << endl;
  }

public:
  PriorityQueue( const GraphType & graph )
  : v()
  , reached()
  , _null(0)
  {
    debug("PriorityQueue()");
  }

  size_t size() const{
    debug("size()");
    return v.size();
  }

  bool empty() const {
    debug("empty()");
    return v.size() == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
    //cout << "push(" << id << "," << key << ")" << endl;
    v.push_back(make_pair(id,key));
    reached[id] = true;
  }

  void deleteMin(){
    //cout << "deleteMin() = " << getMin() << "," << getMinKey() << ")" << endl;
    v.erase(min);
  }

  const id_slot & getMin() {
    //debug("getMin()");
    min = v.begin();
    for (vectorTI it = v.begin(); it < v.end(); it++) {
      if (it->second < min->second) {
        min = it;
      }
    }

    return min->first;
  }

  const key_slot & getMinKey() {
    //debug("getMinKey()");
    return min->second;
  }

  const key_slot & getCurrentKey( id_slot id ) {
    for(vectorTI it = v.begin(); it < v.end(); it++) {
      if (it->first == id) {
        //cout << "getCurrentKey(" << id << ") = " << it->second << endl;
        return it->second;
      }
    }
    //cout << "getCurrentKey(" << id << ") = NULL" << endl;
    return _null;
  }

  bool isReached( id_slot id ) const {
    //cout << "isReached(" << id << ")" << endl;
    return reached.count(id);
  }

  bool contains( id_slot id ) const {
    debug("contains()");
    for(vectorTCI it = v.begin(); it < v.end(); it++)
      if (it->first == id)
        return true;
    return false;
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
    //cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
    for(vectorTI it = v.begin(); it < v.end(); it++)
      if (it->first == id)
        it->second = new_key;
  }

  void increaseKey( id_slot id, key_slot new_key ){
    debug("increaseKey()");
  }

  void updateKey( id_slot id, key_slot new_key ){
    debug("updateKey()");
  }

  void clear(){
    debug("clear()");
    v.clear();
    reached.clear();
  }
};


#endif /* BINARY_HEAP_HPP_ */
